#include <stdio.h>
#include <string.h>
#include <time.h>
#include "common.h"
#include "hash.h"

#define PROG_VERSION "0.2-beta"
#ifndef DEFAULT_TEMPLATE_PATH
#define DEFAULT_TEMPLATE_PATH "/usr/local/lib/lddk/templates"
#endif
#ifndef DEFAULT_INCLUDE_PATH
#define DEFAULT_INCLUDE_PATH  ".:/usr/local/lib/lddk/ddllib"
#endif


#define DEVEL_PATH devel_path
#define TEMPLATE_PATH template_path


int driver_methods=0;
int driver_major = 0;
int driver_minor_to = 0;
int driver_minor = 0;

char devel_path[512];
char template_path[4096];
char include_path[4069];

char driver_objects[1024];
char driver_sources[1024];
char driver_headers[1024];
char lib_objects[1024];
char lib_sources[1024];
char lib_headers[1024];

char submod_includes[1024];
char submod_cases[2046];

/*==========================================================================*/

dump_file(char *dir,char *in, char *out)
{

  FILE *infile;
  FILE *outfile;


  char c,d, *e;
  char buf[180];
  int i=0,j=0;
  int cflag=0;

  char inname[120];
  char outname[120];

  struct hlist_entry *np;


  sprintf(inname,"%s/%s/%s.t",TEMPLATE_PATH,dir,in);
  if( in[strlen(in)-1] == 'c' || in[strlen(in)-1] == 'h' ) {
   sprintf(outname,"%s/%s/%s.%c",DEVEL_PATH,dir,out,in[strlen(in)-1]); cflag=1;
  }
  else {
   sprintf(outname,"%s/%s/%s",DEVEL_PATH,dir,out); cflag=0;
  }

  if(( infile = fopen(inname,"r")) == NULL ){
    printf("\n\n Cant open %s/%s\n Please set your template path correctly\n",TEMPLATE_PATH,in);
    exit(1);
  }

  if(( outfile = fopen(outname,"w")) == NULL ){
    printf("Cant open output %s\n",out);
    exit(1);
  }



  if( cflag )
    fprintf(outfile,"\
 /*
  * Generated with LDDK Version %s 
  *     (c) 1996 Claus Schroeter (clausi@chemie.fu-berlin.de) 
  */
",PROG_VERSION);
#if 0
  else
    fprintf(outfile,"\
#
#  Generated with LDDK Version %s 
#     (c) 1996 Claus Schroeter (clausi@chemie.fu-berlin.de) 
#
",PROG_VERSION);
#endif

  while( (c = fgetc(infile)) != EOF ){  

    if( c == '%' ){
      if(( d = fgetc(infile) ) == EOF ) break;
      if( d == '<' ){
        i=0;
        while(( d = fgetc(infile)) != EOF && d != '>' ){
	  buf[i++] = d;
	}
       	buf[i++] = '\0';
	/*printf("macro '%s'\n",buf);*/
	if(( np = hash_lookup(buf) ) == NULL ){
	  /*printf("\n Warning:unexistent macro %s ignored\n",buf);*/
	}
	else
	  {
	    /*printf("macro %s -> %s\n",buf, np->val );*/
	    fputs( np->val, outfile);
	  }
      } else {
	  fputc(c,outfile);
	  fputc(d,outfile);
	}
    } else {   
    fputc(c,outfile);

    }


  }

  fclose(infile);
  fclose(outfile);

}

/*==========================================================================*/


void add_objects(char *name)
{
  char buf[80];
  sprintf(buf,"%s.o ",name);
  strcat(driver_objects,buf);	
}

void add_sources(char *name)
{
  char buf[80];
  sprintf(buf,"%s.c ",name);
  strcat(driver_sources,buf);	
}

void add_libobjects(char *name)
{
  char buf[80];
  sprintf(buf,"%s.o ",name);
  strcat(lib_objects,buf);	
}

void add_libsources(char *name)
{
  char buf[80];
  sprintf(buf,"%s.c ",name);
  strcat(lib_sources,buf);	
}

void generate_code(char *name)
{

  char buf[1024];
  char buf2[512];
  char buf3[30];
  char *main_name = (char *) hash_get("driver.name");
  int submod=0;

  FILE *out1,*out2;

  /*name = hash_get("driver.name");*/
  if(driver_minor == 0) {
    submod = 0;
    /*hash_install("driver.name",name);*/
  } else {
    submod = 1;
  }
  hash_install("driver.modname",name);

  printf("\n\n Generating Driver '%s' :",name);
     if(submod) printf("(minor=%d)",driver_minor);
  /* 
   * generating file operation struct
   *
   *
   */
  sprintf(buf2,"%s/include/%s_Proto.h",DEVEL_PATH,name);  
  if(( out1 = fopen(buf2,"w")) == NULL ){
	printf("Error: can't open %s \n",buf2);
	exit(1);
  }  


  printf("\n\t fops, ");
  buf[0]='\0';
  sprintf(buf,"struct file_operations %s_fops = { \n", name );
  if( driver_methods & LDDK_SEEK ) {
	sprintf(buf2,"%s_seek, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_seek (struct inode *, struct file *, off_t, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_READ ) {
	sprintf(buf2,"%s_read, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_read (struct inode *, struct file *, char *, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_WRITE ) {
	sprintf(buf2,"%s_write, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_write (struct inode *, struct file *, char *, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_READDIR ) {
	sprintf(buf2,"%s_readdir, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_readdir (struct inode *, struct file *, struct dirent *, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_SELECT ) {
	sprintf(buf2,"%s_select, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_select ( struct inode *, struct file *, int, select_table *); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_IOCTL ) {
	sprintf(buf2,"%s_ioctl, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_ioctl ( struct inode *, struct file *, unsigned int, unsigned long ); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_MMAP ) {
	sprintf(buf2,"%s_seek, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_mmap (struct inode *, struct file *, struct vm_area_struct *); \n",name);
  }  else strcat(buf,"NULL, \n");

  if( driver_methods & LDDK_OPEN_SUB ) {
	sprintf(buf2,"%s_fops_init, \n",main_name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_fops_init (struct inode *, struct file *); \n",main_name);
  } 
  else if( driver_methods & LDDK_OPEN ) {
	sprintf(buf2,"%s_open, \n",name); strcat(buf,buf2);
	fprintf(out1,"extern int %s_open (struct inode *, struct file *); \n",name);
  }  else strcat(buf,"NULL, \n");


  if( driver_methods & LDDK_CLOSE ) {
	sprintf(buf2,"%s_close, \n",name); strcat(buf,buf2);
	fprintf(out1,"void %s_close (struct inode *, struct file *); \n",name);
  }  else strcat(buf,"NULL, \n");

  strcat(buf,"};\n");
  hash_install("driver.fops_struct",buf);
  fflush(out1);
  fclose(out1);

  /* generating prototypes */
  printf("\n\t prototypes, ");


  sprintf(buf2,"#include <%s_Proto.h>",name);
  hash_install("driver.include-proto",buf2);
  sprintf(buf2,"#include <%s_debug.h>",main_name);
  hash_install("driver.include-debug",buf2);
  if( driver_methods & LDDK_IOCTL ){
    sprintf(buf2,"#include <%s_ioctl.h>",name);
    hash_install("driver.include-ioctl",buf2);
    sprintf(buf2,"#include <%s_types.h>",name);
    hash_install("driver.include-types",buf2);
  }

  /*
   * generate files
   *
   */

  printf("\n\t methods, ");
  if( driver_methods & LDDK_SEEK ){
	sprintf(buf2,"%s_seek",name);
	dump_file("driver","seek.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_READ ){
	sprintf(buf2,"%s_read",name);
	dump_file("driver","read.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_WRITE ){
	sprintf(buf2,"%s_write",name);
	dump_file("driver","write.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_READDIR ){
	sprintf(buf2,"%s_readdir",name);
	dump_file("driver","readdir.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_SELECT ){
	sprintf(buf2,"%s_select",name);
	dump_file("driver","select.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_IOCTL ){

        printf("\n\t ioctl functions: ");
        generate_iofuncs(name);

	sprintf(buf2,"%s_ioctl",name);
	dump_file("driver","ioctl.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_MMAP ){
	sprintf(buf2,"%s_mmap",name);
	dump_file("driver","mmap.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_OPEN ){
	sprintf(buf2,"%s_open",name);
	dump_file("driver","open.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_CLOSE ){
	sprintf(buf2,"%s_close",name);
	dump_file("driver","close.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }

  printf("\n\t core, ");fflush(stdout);

  sprintf(buf,"\
    int %s_major = %s_MAJOR; \n\
  ",main_name,main_name);
  hash_install("driver.global-config",buf);

  sprintf(buf,"\
    if( register_chrdev(%s_major,\"%s\",&%s_fops) ){\n\
      printk(\"can't get Major %%d\\n\",%s_major);\n\
      return(-EIO);\n\
    }\n\
  \n",main_name,main_name,name,
      main_name);
  hash_install("driver.register-major",buf);


  sprintf(buf,"\
    if( unregister_chrdev(%s_major,\"%s\") != 0 ){\n\
      printk(\"can't unregister %s, device busy \\n\");\n\
    } else {\n\
      printk(\"%s successfully removed\\n\");\n\
    }\n\
  \n",main_name,main_name,main_name,main_name);
  hash_install("driver.unregister-major",buf);



  if( driver_methods & LDDK_CORE ){
	sprintf(buf2,"%s_core",name);
	dump_file("driver","core.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }




  if( !submod) {
    printf("\n\t headers,");
    hash_install("driver.custom-headers",driver_headers);

    sprintf(buf2,"%s_debug",main_name);
    dump_file("include","debug.h",buf2); add_objects(buf2);
    dump_file("driver","debug.c",buf2); add_sources(buf2);

    sprintf(buf2,"%s_error",main_name);
    /*dump_file("include","error.h",buf2); */
    dump_file("driver","error.c",buf2); add_objects(buf2); add_sources(buf2);

    sprintf(buf2,"%s_P",main_name);
    dump_file("include","P.h",buf2);
  } else {
    printf("\n\t submodule headers, ");
    sprintf(buf2,"%s_fops",name);
    dump_file("include","fops.h",buf2); 

    sprintf(buf,"\n#include <%s_fops.h>" , name ); strcat(submod_includes,buf);
    sprintf(buf,"\n\t case %d:\n\t\t file->f_op = &%s_fops;\n\t\t break;\n\t\t\t" 
    ,driver_minor, name ); strcat(submod_cases,buf);
    

  }
}


do_on_end(){
   char buf2[80];
   char *main_name = hash_get("driver.name");

   hash_install("submod-includes",submod_includes);
   hash_install("submod-cases",submod_cases);

   if( strlen(submod_includes) > 1 ){
  printf("\n\t fops_init");
    sprintf(buf2,"%s_fops_init",main_name);
    dump_file("driver","fops_init.c",buf2); 
    add_sources(buf2); add_objects(buf2);

   }
  printf("\n\t Makefiles");
   hash_install("driver.make-objects",driver_objects);
   hash_install("driver.make-sources",driver_sources);
   dump_file("driver","Makefile","Makefile");


   hash_install("lib.make-objects",lib_objects);
   hash_install("lib.make-sources",lib_sources);
   dump_file("lib","Makefile","Makefile");


   sprintf(buf2,"%s_lib",main_name);
   hash_install("lib.headers",lib_headers);
   dump_file("include","lib.h",buf2);

   dump_file("","Makefile","Makefile");

  printf("\n\n");
}



/*==========================================================================*/

ioctl_func_t driver_iofuncs[MAX_IOFUNCS];
int iofunc_count = 0;

store_iofunct(char *name,char *par,char *code)
{
  if(iofunc_count < MAX_IOFUNCS){
  driver_iofuncs[iofunc_count].name = strsave(name);
  driver_iofuncs[iofunc_count].parlist = strsave(par);
  driver_iofuncs[iofunc_count].code = strsave(code);
  iofunc_count++;
  /*printf("ioctl '%s' (%d) stored\n",name,iofunc_count-1);*/
  } else {
    printf("Error: function capacity exceeded :-( ");
    exit(1);
  }
}


restore_state()
{
  iofunc_count = 0;
  driver_methods = 0;
  driver_minor=0;
  driver_major=0;
  driver_minor_to=0;
}

generate_iofuncs(char *driver_name)
{

 int i,j,k;
 char cases[10*1024];
 char buf[1024],macro[60];
 char *name,*par,*code;
 char *s1,*s2;
 char *argv[10];
 char *main_name = hash_get("driver.name");
 int  argc;


 FILE *codefile,*header,*types,*libfile;

 sprintf(buf,"%s/include/%s_ioctl.h",DEVEL_PATH,driver_name);
   if(( header = fopen(buf,"w") ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }
 sprintf(buf,"%s/include/%s_types.h",DEVEL_PATH,driver_name);
   if(( types = fopen(buf,"w") ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }
 

 strcpy(cases,"");

 for(i=0;i<iofunc_count;i++){
   name = driver_iofuncs[i].name;
   par  = driver_iofuncs[i].parlist;
   code = driver_iofuncs[i].code;

   while( *par == ' ' || *par == '\t' ) par++;
   argc = extract_idents(par,argv);

   printf("\n\t\t %s_%s.c ",driver_name,name);
   sprintf(buf,"%s/driver/%s_%s.c",DEVEL_PATH,driver_name,name);
   if(( codefile = fopen(buf,"w") ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }


   sprintf(buf,"%s_%s",driver_name,name);add_sources(buf); add_objects(buf);   
   fprintf(codefile,"#include <%s_P.h>\n\n\n",main_name);
   fprintf(codefile,"%s_%s( ",driver_name,name);
   fprintf(codefile,"%s",par);
   fprintf(codefile," )\n{\n%s\n}\n",code);
   fclose(codefile);


   sprintf(buf,"%s_%s.c ",driver_name,name);
   strcat(lib_sources,buf);
   sprintf(buf,"%s_%s.o ",driver_name,name);
   strcat(lib_objects,buf);


   strcpy(macro,name);
   fprintf(header,"\n#define %s \t\t %d",uppercase(macro),i);


   sprintf(buf,"%s/lib/%s_%s.c",DEVEL_PATH,driver_name,name);
   if(( libfile = fopen(buf,"w") ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }
   fprintf(libfile,"#include <sys/ioctl.h>\n");   
   fprintf(libfile,"#include <%s_ioctl.h>\n",driver_name);   
   fprintf(libfile,"#include <%s_types.h>\n\n",driver_name);   

   fprintf(libfile,"\
        %s_%s(int fd%c%s)\n\
        {\n\
          extern int errno;\n\
          %s_par_t arg;\n",driver_name,name,(argc>0?',':' '),par,name);

   for(k=0;k<argc;k++){
     /*printf("\n extr:%d=%s",k,argv[k]);*/
     fprintf(libfile,"             arg.%s = %s;\n",argv[k],argv[k]);
   }


   fprintf(libfile,"\t\tioctl(fd,%s,(void *) &arg );\n\
          errno = arg.error;
          return( arg.retval );          
        }\n\n\n\
        ", macro,name);
   fclose(libfile);

   

   fprintf(types,"\ntypedef struct %s_par {\n",name);
   buf[0]='\0';
   for( j=0;j<strlen(par);j++)
     if(par[j] ==',') fprintf(types,";\n");
     else
       fprintf(types,"%c",par[j]);
   fprintf(types,";\n");
   fprintf(types,"int error;\n");
   fprintf(types,"unsigned long retval;\n");
   fprintf(types,"\n} %s_par_t ;\n\n",name);

   sprintf(buf,"      case %s:\n",macro);strcat(cases,buf);
   sprintf(buf,"\
              if( verify_area(VERIFY_READ, (void *) arg, sizeof(%s_par_t))){\n\
                 DBGout(); return(retval); \n\
              }\n\
              if( verify_area(VERIFY_WRITE, (void *) arg, sizeof(%s_par_t))){\n\
                 DBGout(); return(retval); \n\
              }\n\
              argp = (void *) kmalloc( sizeof(%s_par_t) +1 ,GFP_KERNEL );\n\
              memcpy_fromfs( (void *) argp, (%s_par_t *) arg, sizeof(%s_par_t));\n", 
	      name,
	      name,
	      name,
	      name,name );strcat(cases,buf);
   sprintf(buf,"\
              ((%s_par_t *) argp)->retval = %s_%s( ",name,driver_name,name); 
   strcat(cases,buf);

   for(k=0;k<argc;k++){
      sprintf(buf," ((%s_par_t *) argp)->%s",name, argv[k] );strcat(cases,buf);
      if( k+1 < argc ){
	sprintf(buf,", \n\t\t\t\t");strcat(cases,buf);
      }
   }

   sprintf(buf," );\n");strcat(cases,buf); 
   sprintf(buf,"\
              ((%s_par_t *) argp)->error = %s_errno;\n",name,main_name);
   strcat(cases,buf); 
   

   sprintf(buf,"\
              memcpy_tofs( (%s_par_t *) arg, (void *) argp, sizeof(%s_par_t));\n\
              kfree(argp);
              break;\n",
	   name,name);
   strcat(cases,buf);
 }

 hash_install("driver.ioctl-case-functions",cases);
 hash_install("driver.lib-sources",lib_sources);
 hash_install("driver.lib-objects",lib_objects);

  


 fclose(header);
 fclose(types);

}


/*==========================================================================*/

int extract_idents(char *par,char **argv)
{
  int i,j,k;
  char c;
  char *buf;
  char buf2[512];
  int  pargc=0;
  
  buf = par;
  
  i=0;
  while( buf[i] != '\0' && i < strlen(buf) ){
    while( (c=buf[i]) == ' ' || c== '\t' || c == ',' ) i++;
    if( buf[i] == '\0' ) return 0;
    j=0;
    while( (c=buf[i]) != ' ' && c != '\t' && c !='\0' && c != ',' ){
      buf2[j++] = c; i++;
    }
    buf2[j++] = '\0';
    while( (c=buf[i]) == ' ' || c== '\t' ) i++; /* jc */
    if ( buf[i] == ',' || buf[i] == '\0' ){
      for(k=0;k<strlen(buf2);)
	if( buf2[k] == '*' ) k++;
        else 
	  break;
      argv[pargc] = strsave( &buf2[k] );
      /*printf("arg[%d] =%s \n",pargc,argv[pargc]);*/
      pargc++;
    }
  }
  return pargc;

}


/*==========================================================================*/

char *uppercase(char *s)
{
  int i;
  for(i=0;i<strlen(s);i++)
    s[i] = toupper(s[i]);
  
  return s;
}

/*==========================================================================*/

make_sourcetree(){

char *name = hash_get("driver.name");
char cmd[512];

strcpy(devel_path,name);

printf("\n\nGenerating source-tree in ./%s\n",devel_path);fflush(stdout);
sprintf(cmd,"mkdir %s",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/driver",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/include",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/lib",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/doc",devel_path);
system(cmd);

}



/*==========================================================================*/

init_states(){

  strcpy(devel_path,"./tmp");
  strcpy(template_path,DEFAULT_TEMPLATE_PATH);
  strcpy(include_path,DEFAULT_INCLUDE_PATH);

  driver_objects[0]='\0';
  driver_sources[0]='\0';
  driver_headers[0]='\0';
  lib_objects[0]='\0';
  lib_sources[0]='\0';
  lib_headers[0]='\0';
  submod_includes[0] = '\0';
  submod_cases[0] = '\0';
}


init_main(){
  char *date;
  time_t t;

  time(&t);
  date = ctime(&t);
  hash_install("date",date);
  hash_install("version",PROG_VERSION);


}







