%{
#include "hash.h"
#include "y.tab.h"

int yylineno=1;int istate=0;

char codebuf[1024*10];
char *fpath;
char actual_filename[120] = "base";

/* this is for the include statement */


#define MAX_INCLUDE_DEPTH 25

YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int lineno_stack[MAX_INCLUDE_DEPTH];
char *pfile_stack[MAX_INCLUDE_DEPTH];
int include_stack_ptr = 0;

extern char *pfile;
extern char include_path[];

extern char *find_file(char *,char *);
 
%}


%x incl
%x COMMENT
%x DOC
%x CODE
%x STRING
%x PARAMETER

IDENT [A-Za-z_][A-Za-z_0-9]*

%%

"/*"             BEGIN(COMMENT);
<COMMENT>"\n"    { yylineno++; }
<COMMENT>[^*]
<COMMENT>"*"
<COMMENT>"*"+"/" BEGIN(INITIAL);


"/@"		BEGIN(DOC);
<DOC>"\n"       {yylineno++;}
<DOC>[^@]
<DOC>"@"+[^@/]
<DOC>"@"+"/"	BEGIN(INITIAL);


include               BEGIN(incl);
<incl>[ \t]*      /* eat the whitespace */
<incl>[^ \t\n]+   { /* got the include file name */
                   if ( include_stack_ptr >= MAX_INCLUDE_DEPTH )
                       {
                       fprintf( stderr, "Includes nested too deeply" );
                       exit( 1 );
                       }

                   include_stack[include_stack_ptr] =
                       YY_CURRENT_BUFFER;
		   lineno_stack[include_stack_ptr] = yylineno;
		   pfile_stack[include_stack_ptr] = pfile;
                   pfile = strsave(yytext);
		   include_stack_ptr++;	
		   yylineno=1;

		   if(( fpath = find_file(yytext, include_path)) == NULL){
                       fprintf(stderr,"Can't find file: %s from %s\n",yytext,include_path);
		        exit(1);
		        }
                       printf("Loading %s ...\n",fpath);

		       if(( yyin = fopen( fpath, "r" )) == NULL ){
                       fprintf(stderr,"Can't open include file: %s\n",yytext);
		        exit(1);
                       }
                   
                   yy_switch_to_buffer(
                       yy_create_buffer( yyin, YY_BUF_SIZE ) );

                   BEGIN(INITIAL);
		   yylineno++;
                   }
<<EOF>> {
                   if ( --include_stack_ptr < 0 )
                       {
                       yyterminate();
                       }

                   else
                       {
                       yy_delete_buffer( YY_CURRENT_BUFFER );
		       yylineno = lineno_stack[include_stack_ptr];
		       pfile = pfile_stack[include_stack_ptr];
                       yy_switch_to_buffer(
                            include_stack[include_stack_ptr] );
                       }
         }






module                { return T_MODULE; }
driver                { return T_DRIVER; }
library               { return T_LIBRARY; }
methods               { return T_METHODS; }

open                  { return T_OPEN; }
open-submodules       { return T_OPEN_SUB;}
close                 { return T_CLOSE;}
read                  { return T_READ; }
write                 { return T_WRITE; }
mmap                  { return T_MMAP;}
select                { return T_SELECT; }
ioctl                 { return T_IOCTL; }

core                  { return T_CORE; }
init                  { return T_INIT;}
cleanup               { return T_CLEANUP;}
func     	      { return T_FUNCTION; }

name                  { return T_NAME; }
major                 { return T_MAJOR; }
minor                 { return T_MINOR; }
doc                   { return T_DOC;}
dump                  { return T_DUMP;}

user                  { return T_USER;}
file                  { return T_FILE;}
header                { return T_HEADER; }
library               { return T_LIBRARY; }

custom-objects        { return T_COBJECTS; }
custom-sources        { return T_CSOURCES; }
custom-defines        { return T_CDEFINES; }

configurable          { return T_CONFIG; }
macro                 { return T_MACRO; }
proc-template         { return T_TEMPLATE;}


dump                  { return T_DUMP; }

[0-9]+           { sscanf(yytext,"%d",&yylval.ival);   return T_NUMBER; }
0x[0-9a-fA-F]+   { sscanf(yytext,"0x%x",&yylval.ival); return T_NUMBER; }



"/{"             { BEGIN(CODE); codebuf[0] = '\0'; }
<CODE>[^}]*      { macro_strcat(codebuf,yytext); }
<CODE>"}"+[^}/]* { macro_strcat(codebuf,yytext); }
<CODE>"}"+"/"    {
		   /*printf("code = <%s> \n",codebuf);*/
	           BEGIN(INITIAL);
		   yylval.sval = strsave(codebuf);
		   return T_CODE;	
		 }


\"               { BEGIN(STRING); }
<STRING>[^"]*    { yylval.sval = strsave(yytext); 
		   strcat(yylval.sval,"");
		   return T_STRING; }
<STRING>\"       { BEGIN(INITIAL);}

\(		 { BEGIN(PARAMETER); istate=0; }
<PARAMETER>[^)]* { yylval.sval = strsave(yytext); 
		   strcat(yylval.sval,"");
		   istate=1;
		   return T_PARAMETER; }
<PARAMETER>\)    { BEGIN(INITIAL);
		   if(istate == 0) { 
			yylval.sval = strsave(" "); 
			return T_PARAMETER;
		   }
		 }

[a-zA-Z][a-zA-Z_0-9]* { yylval.sval = strsave(yytext); 
			strcat(yylval.sval,"");
			return T_IDENT; }


[ \t]+                      
[ \n ]                        { yylineno++; }

.	         { return(yytext[0]); }

%%

/* do a strcat with macro expansion (for macros in code)*/

macro_strcat(char *t, char *s)
{
  int i=0;
  int j=0;
  int k;
  char c,d,buf[4096],buf2[120];
  extern int yylineno;

  struct hlist_entry *np;

  /*printf("***s='%s'",s);fflush(stdout);*/

  while(( c = s[i++] ) != '\0' ){
	if( c == '%' ){
	  if(( d = s[i++] ) == '\0' ) break;
          if( d == '<' ){
             k=0;
             while(( d = s[i++] ) != '\0' && d != '>' ) buf2[k++] =d;
	     buf2[k++] = '\0';
	     /*printf("buf2=%s\n",buf2);fflush(stdout);*/
	     if(( np = hash_lookup(buf2)) != NULL ){
		for(k=0;k<strlen(np->val);k++) buf[j++]=np->val[k];		
	     } 
	  } else {
	    buf[j++] = c;
            if( d == '\n' ) yylineno++;
            buf[j++] = d;
          }
	} else {
	  if( c == '\n' ) yylineno++;
	  buf[j++] = c;
	}
  }
  buf[j++] = '\0';
  strcat(t,buf);

}

